###############################################################################
#
#    External Jobs, for Odoo
#    Copyright © 2018, 2019, 2022, 2023 XCG Consulting
#                                         (https://www.xcg-consulting.fr/)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################
import datetime
import logging
import os
import re
from base64 import b64encode
from typing import Any

import requests
from odoo import _, api, fields, models  # type: ignore[untyped-import]
from odoo.tools import config  # type: ignore[untyped-import]

from .job_log import State

_logger = logging.getLogger(__name__)

_TIMEOUT = os.environ.get("ODOO_ADDONS_EXTERNAL_JOB_TIMEOUT", None)
ODOO_ADDONS_EXTERNAL_JOB_TIMEOUT = None if _TIMEOUT is None else float(_TIMEOUT)


class JobDefinition(models.Model):
    """Represents a job definition."""

    _name = "external_job.job_definition"
    _description = "External job definition"

    name = fields.Char(string="Name", size=128, required=True)

    short_desc = fields.Char(string="Short Description", size=256, translate=True)

    long_desc = fields.Text(string="Long Description")

    output_filename_base = fields.Char(
        string="Output File Name",
        size=128,
        help=(
            "The desired file pattern . For example:"
            " %(job_definition.name)s-%(datetime.today)s.csv. If there is no"
            " value, no output file is created and reply is added to job log stdout."
        ),
        groups="base.group_system",
    )

    model_id = fields.Many2one(
        comodel_name="ir.model",
        string="Model",
        ondelete="cascade",
        help=(
            "The data model our job is expecting to process, used in the "
            "job runner wizard."
        ),
    )

    force_commit = fields.Boolean(
        string="Force Commit",
        help=(
            "Only needed when the external job need to see the job log "
            "in the database. Avoid as much as possible."
        ),
        default=False,
        groups="base.group_system",
    )

    job_type = fields.Selection(
        selection=[("extrunner", "Extrunner")],
        string="Job Type",
        help="Choose the way to run the job.\n"
        " Extrunner: Call an external service to launch the job.\n"
        " Command line: Run the external process from odoo.",
        default="extrunner",
        groups="base.group_system",
    )

    extrunner_server_id = fields.Many2one(
        "external_job.extrunner_server",
        string="Extrunner Server",
        ondelete="restrict",
        groups="base.group_system",
    )

    job_ID = fields.Integer(
        string="Job ID",
        help="ID of the job to be launched by extrunner.",
        groups="base.group_system",
    )

    env_var_ids = fields.One2many(
        comodel_name="external_job.job_definition.env_var",
        inverse_name="job_def_id",
        string="Environment variables",
        help=(
            "Variables sent to extrunner. May contain dynamic arguments (see "
            "legend at the bottom of the job definition form)."
        ),
        groups="base.group_system",
    )

    default_ids = fields.One2many(
        comodel_name="external_job.job_definition.default_value",
        string="Defaults",
        help="Dictionary to use as defaults (only implemented for context values)",
        inverse_name="job_def_id",
        groups="base.group_system",
    )

    job_ids = fields.One2many(
        comodel_name="external_job.job_log",
        inverse_name="job_definition_id",
        string="Jobs",
    )

    job_count = fields.Integer(
        "# Jobs",
        help="Number of jobs using that definition",
        compute="_compute_job_count",
        compute_sudo=True,
    )

    unlink_delay = fields.Integer(
        "Unlink Delay",
        help="Job logs that ended less than this delay in hours will not be unlinked. "
        "If not set, no unlink is done.",
    )

    _sql_constraints = [
        (
            "check_unlink_delay",
            "check(unlink_delay >= 0 or unlink_delay is null)",
            "Unlink delay must be positive or unset!",
        )
    ]

    @api.depends("job_ids")
    def _compute_job_count(self):
        for definition in self:
            definition.job_count = len(definition.job_ids)

    def action_run_job(self):
        """This callback method is called by a user pressing a button in the form view.
        When the user presses the button we run the command and collect the data back,
        then display a view of the job."""
        job_logs = self.run_job()
        if job_logs:
            return self.return_joblog_act_window(job_logs)
        return True

    @api.returns("external_job.job_log")
    def run_job(self) -> models.BaseModel:
        """Run jobs

        :return: external_job.job_log records, to get a view change, use
            return_joblog_act_window, or use action_run_job directly
        """
        # User do not have the rights to see most of the fields, or in the case of
        # extrunner, the models used, so use sudo to bypass the restrictions
        joblogs = (
            self.sudo().env["external_job.job_log"].create(self.sudo()._vals_list())
        )
        self.sudo()._run(joblogs)
        return joblogs.sudo(False)

    def _vals_list(self) -> list[dict[str, Any]]:
        vals_list = []
        extrunner_jobs = self.filtered(lambda job: job.job_type == "extrunner")
        # Pylint does not understand that this is ourselves
        # pylint: disable=protected-access
        if extrunner_jobs:
            vals_list.extend(extrunner_jobs._vals_list_extrunner())
        return vals_list

    def _run(self, joblogs: models.BaseModel) -> None:
        extrunner_job_logs = joblogs.filtered(
            lambda joblog: joblog.job_definition_id.job_type == "extrunner"
        )
        # Pylint does not understand that this is ourselves
        # pylint: disable=protected-access
        if extrunner_job_logs:
            self._call_extrunner(joblogs)

    @staticmethod
    def return_joblog_act_window(joblogs: models.BaseModel):
        """Return a view change from a list of job log ids"""
        if len(joblogs) == 1:
            return JobDefinition._return_joblog_act_window(joblogs, default_view="form")
        return JobDefinition._return_joblog_act_window(joblogs, view_mode="tree")

    @staticmethod
    def _return_joblog_act_window(
        job_logs: models.BaseModel, view_mode="form", default_view="tree"
    ):
        """return a dictionary that represent external job log action
        Note: default_view : if default view is form, specifies the
        record to load.
        """
        act = {
            "name": _("Job Log"),
            "type": "ir.actions.act_window",
            "res_model": "external_job.job_log",
            "view_mode": view_mode,
        }

        if len(job_logs) == 1 and default_view == "form":
            act["res_id"] = job_logs[0].id
            return act

        act["domain"] = [("id", "in", job_logs.ids)]
        return act

    def _vals_list_extrunner(self) -> list[dict[str, Any]]:
        return [{"job_definition_id": rec.id} for rec in self]

    @api.model
    def _call_extrunner(self, job_logs: models.BaseModel) -> None:
        """Call extrunner to launch an external job."""
        for joblog in job_logs:
            rec = joblog.job_definition_id

            # Dynamic arguments: Similar to what we do in the "run_job" method;
            # not worth factoring out.
            arg_dict = ArgumentDict(
                rec,
                joblog,
                rec.env.context,
                rec.env.cr,
                defaults=rec.default_ids.to_dict(),
            )

            # login
            headers = {"Content-Type": "application/json"}
            server = rec.extrunner_server_id
            post = requests.post(
                f"{server.base_url}/auth/login",
                json={
                    "login": server.user_login,
                    "password": server.user_password,
                },
                headers=headers,
                timeout=ODOO_ADDONS_EXTERNAL_JOB_TIMEOUT,
            )
            if post.status_code != 200:
                JobDefinition._handle_error(
                    joblog,
                    _(
                        "status code: %(code)s\nFailed to login to extrunner: %(text)s",
                        code=post.status_code,
                        text=post.text,
                    ),
                )
                return

            headers["Authorization"] = f"Bearer {post.json().get('id_token', 'failed')}"

            # Make sure a commit of our transaction before giving the newly
            # created id to our external process, or it won't be able to find it
            if rec.force_commit:
                rec.env.cr.commit()

            # Include our env var key/value pairs into the JSON data sent here.
            # Values support dynamic formatting via "ArgumentDict".
            data = {"env": {var.key: var.value % arg_dict for var in rec.env_var_ids}}
            data["env"].update({"dbname": rec.env.cr.dbname})

            args = rec.env.context.get("args")
            if args:
                for arg in re.split("--", args):
                    res = arg.split()
                    if res:
                        data["env"].update({res[0]: res[1]})

            post = requests.post(
                f"{server.base_url}/run/{rec.job_ID}",
                headers=headers,
                json=data,
                timeout=ODOO_ADDONS_EXTERNAL_JOB_TIMEOUT,
            )

            if post.status_code != 200:
                if post.status_code == 400:
                    # Bad request
                    JobDefinition._handle_error(
                        joblog,
                        _(
                            "Bad status code: %(code)s\nBad request: %(text)s",
                            code=post.status_code,
                            text=post.text,
                        ),
                    )
                if post.status_code == 401:
                    # Unauthorized
                    JobDefinition._handle_error(
                        joblog,
                        _("Bad status code: %s\nUnauthorized", post.status_code),
                    )
                if post.status_code == 520:
                    # Error server
                    JobDefinition._handle_error(
                        joblog,
                        _(
                            "Bad status code: %(code)s\n"
                            "Server internal error: %(text)s",
                            code=post.status_code,
                            text=post.text,
                        ),
                    )
                continue

            values = {"end_date": fields.Datetime.now(), "state": State.DONE.value}
            if rec.output_filename_base:
                values.update(
                    {
                        "filename": rec.output_filename_base % arg_dict,
                        "out_file": b64encode(post.content),
                    }
                )
            elif post.content:
                values["stdout"] = post.content.decode("UTF-8")

            joblog.write(values)

    @staticmethod
    def _handle_error(joblog: models.BaseModel, msg: str):
        joblog.write(
            {
                "end_date": fields.Datetime.now(),
                "state": "exception",
                "stderr": msg,
            }
        )

    def action_show_jobs(self):
        """Action to show jobs linked to a job definition"""
        self.ensure_one()
        return {
            "name": _("Jobs"),
            "view_mode": "tree,form",
            "res_model": "external_job.job_log",
            "type": "ir.actions.act_window",
            "context": {"create": False, "delete": False},
            "domain": [("id", "in", self.job_ids.ids)],
            "target": "current",
        }

    @api.autovacuum
    @api.model
    def _gc_job_log(self):
        """Unlink old done job logs"""
        for job_definition in self.search([("unlink_delay", "!=", False)]):
            self.env["external_job.job_log"].search(
                [
                    ("job_definition_id", "=", job_definition.id),
                    ("state", "=", State.DONE.value),
                    ("end_date", "!=", False),
                    (
                        "end_date",
                        "<",
                        fields.Datetime.now()
                        - datetime.timedelta(hours=job_definition.unlink_delay),
                    ),
                ]
            ).unlink()


class ArgumentDict:
    """Used to create arguments of a job."""

    def __init__(self, job_definition, joblog, context, cr, defaults=None):
        self.job_definition = job_definition
        self.joblog = joblog
        self.context = context
        self.cr = cr
        self.defaults = {} if defaults is None else defaults

    # incorrect but does not seem to be used
    def __len__(self):
        return 3 + len(self.context)

    def __getitem__(self, key):
        parts = key.split(".")
        if parts[0] == "job_log":
            # dangerous?
            return getattr(self.joblog, parts[1])
        if parts[0] == "job_definition":
            # dangerous?
            return getattr(self.job_definition, parts[1])
        if parts[0] == "context":
            if self.context:
                if parts[1] in self.context:
                    return self.context[parts[1]]
                return self.defaults[key]
        if parts[0] == "config":
            if parts[1] == "options":
                if parts[2] == "db_port":
                    return (
                        config.options[parts[2]] if config.options[parts[2]] else 5432
                    )
                return config.options[parts[2]]
            return config.misc.get(parts[1], {}).get(parts[2])
        if parts[0] == "cr":
            if parts[1] == "dbname":
                return self.cr.dbname
        if parts[0] == "datetime":
            if parts[1] == "now":
                return fields.Datetime.now()
            if parts[1] == "today":
                return fields.Date.today()
        raise KeyError

    # does not seem to be used
    def __contains__(self, item):
        if item in [
            "job_log.id",
            "job_log.in_file_name",
            "job_log.out_file_name",
        ]:
            return True
        parts = item.split(".")
        if parts[0] == "job_definition":
            # TODO do real test
            return True
        if parts[0] == "context":
            return self.context and (parts[1] in self.context or item in self.defaults)
        if parts[0] == "config":
            if parts[1] == "options":
                return parts[2] in config.options
            return parts[2] in config.misc.get(parts[1], {})
        if parts[0] == "cr":
            if parts[1] == "dbname":
                return True
        return False

    # XXX __iter__ not used apparently
