###############################################################################
#
#    External Jobs, for Odoo
#    Copyright © 2014-2016, 2022-2024 XCG Consulting (https://www.xcg-consulting.fr/)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################
import datetime
import logging
from enum import Enum

from odoo import _, api, fields, models  # type: ignore[untyped-import]

_logger = logging.getLogger(__name__)


class State(Enum):
    STARTED = "started"
    DONE = "done"
    EXCEPTION = "exception"

    def translate(self):
        match self:
            case State.STARTED:
                return "Started"
            case State.EXCEPTION:
                return "Exception"
            case State.DONE:
                return "Done"


class JobLog(models.Model):
    """The log of an external job"""

    _name = "external_job.job_log"
    _description = "External Job Log"

    job_definition_id = fields.Many2one(
        comodel_name="external_job.job_definition",
        string="Job Definition",
        ondelete="cascade",
        required=True,
        index="btree_not_null",
    )

    start_date = fields.Datetime(
        string="Start Date", readonly=True, default=fields.Datetime.now, copy=False
    )

    # TODO rename to end_datetime?
    end_date = fields.Datetime(string="End Date", readonly=True, copy=False)

    in_file_name = fields.Char(
        string="Temporary In File Name", size=512, readonly=True, copy=False
    )

    out_file_name = fields.Char(
        string="Temporary Out File Name", size=512, readonly=True, copy=False
    )

    out_file = fields.Binary(string="Output File", readonly=True, copy=False)

    filename = fields.Char(string="Filename", copy=False)

    stdout = fields.Text(string="StdOut", readonly=True, copy=False)

    state = fields.Selection(
        selection=[(state.value, state.translate()) for state in State],
        string="State",
        default=State.STARTED.value,
        required=True,
        index="btree_not_null",
    )

    @api.depends("start_date")
    @api.depends_context("lang")
    def _compute_display_name(self):
        """Redefined to use the start_date as the name but using the locale lang and
        keeping tz as per context info"""
        lang_model = self.env["res.lang"]
        for jlog in self:
            lang = lang_model.search([("code", "=", self.env.lang)])
            if lang:
                lang_format = f"{lang.date_format} {lang.time_format}"
            else:
                # ISO format
                lang_format = "%Y-%m-%dT%H:%M:%S"

            job_dtime = fields.Datetime.from_string(jlog.start_date)

            job_dtime = fields.Datetime.context_timestamp(jlog, job_dtime)

            job_dtime = datetime.datetime.strftime(job_dtime, lang_format)

            jlog.display_name = _("%s’s Job", job_dtime)

    @api.returns("self", lambda value: value.id)
    def copy(self, default=None):
        vals = {}
        if default is not None:
            vals.update(default)
        vals.update(self.job_definition_id._vals_list()[0])
        created = super().copy(vals)
        created.job_definition_id._run(created)
        return created
