###############################################################################
#
#    External Job, for Odoo
#    Copyright © 2018, 2022, 2023, 2024 XCG Consulting (https://www.xcg-consulting.fr/)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

import base64
from unittest import mock

import requests  # Odoo req
from freezegun import freeze_time
from odoo.exceptions import AccessError  # type: ignore[untyped-import]

from ._common import ExtrunnerTest


class Test(ExtrunnerTest):
    """Test external job definition.

    These tests assume an Odoo database with demo data
    """

    def test_run_job_echo_1(self):
        """Run a job and assert that the result id is in the output if no file is
        selected.
        Also test gc."""
        job_def = self.job_definition
        job_def.output_filename_base = None
        self.__class__.extrunner_request_content = b"1"
        with freeze_time("2023-02-01 10:30:00"):
            job_log = job_def.with_user(self.user_demo).run_job()

        self.assertEqual("1", job_log.stdout)

        self.assertEqual(job_def.job_count, 1)
        self.assertEqual(job_log.state, "done")
        # gc tests
        job_def.unlink_delay = 1
        self.env["external_job.job_definition"]._gc_job_log()
        self.assertEqual(
            job_def.job_count, 0, "autovacuum should delete the old job log"
        )

    def test_run_job_echo_2(self):
        """Run a job and assert that the output is correct.
        Also test gc not done if not set."""
        job_def = self.job_definition
        job_def.output_filename_base = None
        self.__class__.extrunner_request_content = b"echo 2\n"
        with freeze_time("2021-02-02 17:30:00"):
            job_log = job_def.with_user(self.user_demo).run_job()
        self.assertEqual("echo 2\n", job_log.stdout)
        self.assertEqual(job_def.job_count, 1)
        self.env["external_job.job_definition"]._gc_job_log()
        self.assertEqual(job_def.job_count, 1, "no unlink should be done if delay is 0")

    @mock.patch("requests.post")
    def test_run_job_extrunner(self, requests_post_mock):
        """Run an extrunner job."""

        # Mock the HTTP POST request.
        requests_post_mock.return_value = mock.Mock(
            content=b"ATTACHMENT-CONTENTS", status_code=requests.codes.ok
        )

        # Security tests
        with self.assertRaises(AccessError):
            self.assertEqual(
                self.env["external_job.extrunner_server"]
                .with_user(self.user_demo)
                .search_count([]),
                0,
            )

        with self.assertRaises(AccessError):
            self.server.with_user(self.user_demo).user_password  # noqa: B018 # pylint: disable=expression-not-assigned
        # Created extrunner job def.
        job_def = self.job_definition

        # Run it!
        self.assertEqual(job_def.job_count, 0)  # no jobs yet.
        job = job_def.with_user(self.user_demo).run_job()
        self.assertEqual(len(job), 1)  # 1 new job.

        # Our POST mock should have been called twice (auth & run).
        requests_post_mock.assert_any_call(
            "http://BASE-URL-1/auth/login",
            json=mock.ANY,
            headers=mock.ANY,
            timeout=mock.ANY,
        )
        requests_post_mock.assert_any_call(
            "http://BASE-URL-1/run/0", json=mock.ANY, headers=mock.ANY, timeout=mock.ANY
        )

        # Check key/value env vars sent in the last call. See job_def creation
        # above for env vars we test here.
        last_call_data = requests_post_mock.call_args[-1]["json"]
        self.assertEqual(
            last_call_data,
            {
                "env": {
                    "dbname": self.env.cr.dbname,
                    "STATIC-1": "1",
                    "DYN-1": "DEF-1 x",  # jobdef name & " x"
                }
            },
        )

        # Check the new job.
        self.assertTrue(job.end_date)
        self.assertEqual(job.state, "done")
        # Ensure the new job has a file with contents from the mock.
        self.assertEqual(job.out_file, base64.b64encode(b"ATTACHMENT-CONTENTS"))
        self.assertTrue(job.filename)

        # copy tests
        # Before 2.0.0, the test used id=1 for tests and the copy would work.
        copy = job.sudo().copy()
        requests_post_mock.assert_any_call(
            "http://BASE-URL-1/auth/login",
            json=mock.ANY,
            headers=mock.ANY,
            timeout=mock.ANY,
        )
        requests_post_mock.assert_any_call(
            "http://BASE-URL-1/run/0", json=mock.ANY, headers=mock.ANY, timeout=mock.ANY
        )
        self.assertTrue(copy.end_date)
        self.assertEqual(copy.state, "done")
        self.assertEqual(copy.out_file, base64.b64encode(b"ATTACHMENT-CONTENTS"))
        self.assertTrue(copy.filename)

        # Added to ensure users can not duplicate logs
        with self.assertRaises(AccessError):
            job.copy()
